!******************************************************************************
!
! File : relaxsys.f90
!   by : Alan Tackett
!   on : 7/27/98
!  for : PAW project
!
!  This module contains the routines for relaxing the electrons and 
!  also the atoms.  
!
!******************************************************************************

Module relaxsys

Use paw_inout
Use misc
Use options_data
Use word
Use atom_data
Use crystal_data
Use crystal_symmetry
Use psilib
Use solver
Use mem_data
Use memmgr
Use forces
Use work_mgr
Use storedata
Use timedependent
Use anderson_mixing
Use gradenk

Implicit NONE

Integer, PARAMETER :: RELAX_MD       = 0
Integer, PARAMETER :: RELAX_GEOMETRY = 1

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  CheckAtomOverlap - Checks to make sure that none of the atomic spheres
!      overlap
!
!******************************************************************************

Subroutine CheckAtomOverlap
  Integer :: i, j, k, a1, a2
  Real    :: DX(3), R(3), Rc, R0(3), T(3), Rct, r2

  Do a1=1, Specific_Atoms
     Rc = AtomType_Info(Atom_List(a1)%TypeIndex)%Rc
     R0 = Atom_List(a1)%Pos

     Do i=-1, 1
        Do j=-1, 1
           Do k=-1, 1
              T = MatMul(xtal%Basis, (/i,j,k/))
              Do a2=a1+1, Specific_Atoms  
                 Rct = Rc + AtomType_Info(Atom_List(a2)%TypeIndex)%Rc
                 R = R0 - (Atom_List(a2)%Pos + T)
                 r2 = SQRT(DOT_PRODUCT(R, R))
                 If (r2 < RcT) then
                    Write(Error_Unit,*) 'CAO: A1=',&
                         TRIM(Atom_List(a1)%Name), ' overlaps ', &
                         TRIM(Atom_List(a2)%Name), ' * Rc=',Rc,  &
                         ' * RcT=',RcT,' * R=', &
                         R2, ' * ijk=',i,j,k
                 End If
              End Do
           End Do
        End Do
     End Do
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  UpdateSystem - Performs system update each time the atoms move.
!
!******************************************************************************

Subroutine UpdateSystem

  Integer :: i,j,Lmax

  If (xtal%Auto_Symmetry) then
      !******* Initialize the Crystal_Symmetry routines **********
     Lmax = -1
     Do i=1, Atom_Types
        j = MaxVal(AtomType_Info(i)%L_Value)
        Lmax = MAX(Lmax, j)
     End Do
     
     Call StoreSymmetry('RUNTIME')
     If (xtal%Rot_Size > 0) Call SetSym(Lmax)
  End If


  Call LRU_FreeAll(LRU_Phase)
  Call StoreStructFactors
  Call CalcVlocal(SCFvalues%V_local)

  If (Run_Mode == MIN_TIME) then
     Do i=1, Specific_Atoms
        Call LRU_FreeAll(LRU_Proj(i))
        Close(LRU_Proj(i)%FD_Base)
     End Do
     Call StoreRawProj
  end If

  If (Proj_Mode == PROJ_RS)  Call RS_Update

  Call AllocPrecVec  !** Allocate space for the JDQZ preconditioner

  Return
End Subroutine


!******************************************************************************
!
!  UpdateKpoint - Updates the system set of K-points for band structure
!                 calculations
!
!    NewK - New list of kpoints should be the same size as BZ%Ku.
!
!******************************************************************************

Subroutine UpdateKpoint(NewK)
   Real, Intent(IN) :: NewK(:)

   Integer :: i

   BZ%Ku(:,1) = NewK  !** Assign New Kpoint set


   PsiInfo(:)%PDOT_Stored = .FALSE.
   PsiInfo(:)%KE_Stored = .FALSE.
 
Write(Log_Unit,*) 'UpadteKpoint K1=',NewK(:) ; call flush_(Log_unit)
   Do i=1, Atom_Types
      Call LRU_FreeAll(LRU_RadProj(i))
      Close(LRU_RadProj(i)%FD_Base)
   End Do

   Call StoreRadProj

   Call LRU_FreeAll(LRU_Ylm)
   Call StoreYlm

   Call UpdateSystem
RETURN

   !*** Purge the Buffers ***
   Do i=1, Atom_Types
      Call LRU_FreeAll(LRU_RadProj(i))
      Close(LRU_RadProj(i)%FD_Base)
   end Do

   Call LRU_FreeAll(LRU_Phase)
   Call LRU_FreeAll(LRU_Ylm)

   Do i=1, Specific_Atoms
      Call LRU_FreeAll(LRU_Proj(i))
      Close(LRU_Proj(i)%FD_Base)
   End Do

   Call StoreRadProj         !** Update radial projectors
   Call StoreStructFactors   !** Update Structure Factors
   Call StoreYlm             !** Update the Ylm's

   If (Run_Mode == MIN_TIME) Call StoreRawProj

End Subroutine

!******************************************************************************
!
!  PrintOccAndEig - Prints to the Log_Unit the Old, new, and diff 
!                   occupancies, eigenvalues, and fermi energy
!
!  OldOcc   - Old occupancies
!  OldEn    - Old energy Eigenvalues
!  OldFermi - Old Fermi energy level
!
!******************************************************************************

Subroutine PrintOccAndEig(OldOcc, OldEn, OldFermi)
  Real, Intent(IN) :: OldOcc(:,:)
  Real, Intent(IN) :: OldEn(:,:)
  Real, Intent(IN) :: OldFermi

  Integer :: i,j,k
  Real    :: docc, den, df
  Character*100 :: Msg

  Write(Log_Unit, *) '----------------------------------------------'
  Write(Log_Unit, *) 'New Fermi:',BZ%Fermi, ' * Old Fermi:',OldFermi, &
                     'Diff:',BZ%Fermi-OldFermi

  msg = '(2i5," ",3e15.7)'
  Write(Log_Unit, *) 'Occupancy-----------------------------------------------'
  Write(Log_Unit,*) 'k-pnt  Band     NewOcc         OldOcc        DiffOcc'
  Do k=1, NumKpnts
     Do i=1, NumBands
        Write(Log_Unit, msg) k,i, Occupancy(i,k), OldOcc(i,k), &
             (Occupancy(i,k)-OldOcc(i,k))
     End Do
  End Do

  msg = '(2i5," ",3e15.7)'
  Write(Log_Unit, *) 'Energy Levels------------------------------------------'
  Write(Log_Unit,*) 'k-pnt  Band     NewEig         OldEig        DiffEig'
  Do k=1, NumKpnts
     Do i=1, NumBands
        Write(Log_Unit,msg) k, i, BZ%Keu(i,k), OldEn(i,k), &
             (BZ%Keu(i,k)-OldEn(i,k))
     End Do
  End Do

  Write(Log_Unit, *) '-----------------------------------------------------'
  Return
End Subroutine

!******************************************************************************
!
!  RelaxElectrons - Relax's the wave functions until equilibrium is met or 
!                   the maximum number of iterations is reached.
!
!  MaxIter - Max number of SCF iterations
!  Tol     - Error tolerance for the Energy
!
!******************************************************************************

Subroutine RelaxElectrons(MaxIter, Tol, Converged)
  Integer,       Intent(IN) :: MaxIter
  Real,          Intent(IN) :: Tol
  Logical,       Intent(OUT)       :: Converged

  Integer       :: Iter, i,j, Lmax
  Character*100 :: msg
  Real  ::  Etol

  If (.not.bandstructure_mode) then
  Call Printdate(Output_Unit,'#  Starting SCF on date ')
  Write(Output_Unit,*) ' '
  Write(Output_Unit,*) '# Current Fractional Positions '
  Write(Output_Unit,*) '#Atom_List FRAC_POSITION'
  Do i=1, Specific_Atoms
     j=Atom_List(i)%TypeIndex
     If (Atom_List(i)%Freeze) then
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7, " # FROZEN")') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Frac_Pos
     Else If (Anchor.and.i==AnchorIndex) then
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7, " # ANCHORED")') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Frac_Pos
     else
        write(Output_Unit,'(" # ", A8,"  ",A8, 1p3e15.7)') &
             Atom_List(i)%Name, &
             AtomType_Info(j)%Atom_Name, &
             Atom_List(i)%Frac_Pos
     EndIf
   Enddo
   Call Flush_(Output_Unit)
  EndIf

  Call CheckAtomOverlap
  CohesiveEnergy=1E30  !***REMOVE******
  Prev_CohesiveEnergy = 1E30

  Etol = Tol
  Call SCF_Solver( MaxIter, ETol, iter)

  Converged = .FALSE.
  If (ETol < Tol) Converged = .TRUE.

  Call Printdate(Output_Unit,'#  SCF results generated on date ')

  If (Converged) then
     Write(Log_Unit, *) 'RelaxElectrons: Converged in ',iter, &
          'iterations with a final Energy error of ', Etol
     Write(Output_Unit, *) 'RelaxElectrons: Converged in ',iter, &
          'iterations with a final Energy error of ', Etol
  else
     Write(Log_Unit, *) 'RelaxElectrons: Did Not converged in ',iter, &
          'iterations with a final energy error of ', Etol
     Write(Output_Unit, *) 'RelaxElectrons: Did Not converged in ',iter, &
          'iterations with a final energy error of ', Etol
  End If

  Write(Output_Unit,*) 'RelaxElectrons: Cohesive Energy:',CohesiveEnergy
  Write(Output_Unit,*) 'RelaxElectrons: # iterations   :',iter ,'/',MaxIter
  Write(Output_Unit,*) 'RelaxElectrons: Tolerance      :',Etol ,'/',Tol
  Write(Output_Unit,*) ' '
  Call DisplayEnergy(Output_Unit)
  Call Flush_(Output_Unit)

  If (.NOT. Converged) then
     Write(Msg, *) 'RelaxElectrons: Convergence NOT satisfied but exiting!'
     Write(Error_Unit, *) Msg
     IF (Error_Unit /= Log_Unit) Write(Log_Unit, *) Msg
     If ((Error_Unit /= Output_Unit) .AND. (Log_Unit /= Output_Unit)) then
        Write(Error_Unit, *) Msg
     End IF
  End If

  Return
End Subroutine


!******************************************************************************
!
!   MakeGeometryMixing - Initializes the Geometry relaxation data structure
!
!******************************************************************************

Subroutine MakeGeometryMixing
  Integer :: i, j, err, s,t
  Character*100 :: Msg
  
  j = 0  !*** First determine how many atoms are allowed to move
  Do i=1, Specific_Atoms
     If (.NOT. Atom_List(i)%Freeze) j = j + 1
  End Do

  If (Associated(Move_AC)) then
     Call FreeAnderson(Move_AC)
  End If

  i = 3*j
  Call InitAnderson(Move_AC, Error_Unit, Mix_Size, i, 1E-5, Geometry_NewMix)

  Return
End Subroutine

!******************************************************************************
!
!  GeometryMixingStep - Updates the atom positions
!
!******************************************************************************

Subroutine GeometryMixingStep
  Integer :: i, j, k
  Complex :: OldPos(3*Specific_Atoms), DiffPos(3*Specific_Atoms)
  Real :: mag
  
  j = 1
  Do i=1, Specific_Atoms
    If (.NOT. Atom_List(i)%Freeze) then
       OldPos(j:j+2) = Atom_List(i)%Pos
       Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Force%Pos(:,1) + &
            Atom_List(i)%Force%CurrForce
       DiffPos(j:j+2) = Atom_List(i)%Force%Pos(:,1)  !** New Position

       j = j + 3
    End If
  End Do

  j = j - 1
  DiffPos(1:j) = DiffPos(1:j) - OldPos(1:j)
  Call Anderson_Mix(Move_AC, OldPos(1:j), DiffPos(1:j))

  j = 1
  Do i=1, Specific_Atoms
    If (.NOT. Atom_List(i)%Freeze) then
       Atom_List(i)%Force%Pos(:,1) = OldPos(j:j+2)  !** Now OldPos=MixedPos

       j = j + 3
    End If
  End Do

  Return
End Subroutine


!******************************************************************************
!
! MoveAtom - Calculates the new atom position given the Verlet data 
!
! Force - Verlet data structure containg the pos, Vel, and acc information 
!         for the atom
! Mass  - Atoms mass
! Friction - Friction coeff
!
!******************************************************************************

Subroutine MoveAtom(Force, mass, Friction)
  Type (Verlet_Data), Intent(INOUT) :: Force
  Real,       Intent(IN)    :: Mass
  Real,       Intent(IN)    :: Friction

  Integer :: i, j
  Real :: DT, DT2

  DT = TimeStep
  DT2 = DT*DT

  j = Force%NumMoves    !*** Copy R,V, and A to make room for new data
  If (j>2) j = 2
  Do i=j, 1, -1
     Force%Pos(:,i+1) = Force%Pos(:,i)
     Force%Velocity(:,i+1) = Force%Velocity(:,i)
     Force%Acceleration(:,i+1) = Force%Acceleration(:,i)
  End Do

  Force%Acceleration(:,1) = Force%CurrForce/Mass - Friction*Force%Velocity(:,2)

  Where (ABS(Force%Acceleration) < Force_Zero) Force%Acceleration = 0

  If (Force%NumMoves < 3) then  !*** First few steps are Euler steps
     Force%Pos(:,1) = Force%Pos(:,2) + Force%Velocity(:,2)*DT + &
          0.5*Force%Acceleration(:,1)*DT2
     Force%Velocity(:,1) = Force%Velocity(:,2) + Force%Acceleration(:,1)*DT
     Force%NumMoves = Force%NumMoves + 1
     Force%Predictor_Phase = .TRUE.

  else           !**** Can now take Beeman-Verlet steps *****
     Force%Pos(:,1) = 2*Force%Pos(:,2) - Force%Pos(:,3) + &
          Force%Acceleration(:,1)*DT2
     Force%Velocity(:,1) = (Force%Pos(:,1) - Force%Pos(:,2))/DT

!     Force%Velocity(:,1) = Force%Pos(:,1) - Force%Pos(:,2) + &
!          (2*Force%Acceleration(:,1)+Force%Acceleration(:,2))*DT2/6.0
!     Force%Velocity(:,1) = Force%Velocity(:,1)/DT
!     
!     Force%Pos(:,1) = Force%Pos(:,2) + DT*Force%Velocity(:,1) + &
!          (4*Force%Acceleration(:,2)-Force%Acceleration(:,3))*DT2/6.0
  End If

  Return
End Subroutine


!******************************************************************************
!
!  MoveAllAtoms - Moves the atoms using the standard Beeman-Verlet method
!
!  Mode - Type of move to perform, ie, MD or Broyden step for geometry
!
!******************************************************************************

Subroutine MoveAllAtoms(Mode)
  Integer, Intent(IN) :: Mode

  Real :: CurrForce(3), FrictionForce(3)
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Integer :: i, j

  Write(Log_Unit,*) 'MoveAllAtoms:  Printing Updated Atom Positions and Forces'

  !If (Mode == RELAX_GEOMETRY) Call GeometryMixingStep -- now done in relaxatoms

  Do i=1, Specific_Atoms 
           !** Calculate the new position *
     If (Mode == RELAX_MD) Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Pos 

     Write(Log_Unit, &
          '("MoveAllAtoms: DT = ",1e10.2, " * Friction = ", 1e10.2, " * Mass = ",1e10.2)') &
          TimeStep, Atom_List(i)%Friction, &
          Atomtype_Info(Atom_List(i)%TypeIndex)%Mass

!Write(Log_Unit,*) 'MoveAllAtoms: DT=',TimeStep, ' * Friction=',&
!     Atom_List(i)%Friction, ' * Mass=', &
!     Atomtype_Info(Atom_List(i)%TypeIndex)%Mass

     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos
     Write(Log_Unit,'("MoveAllAtoms: ", A10," F(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Force%CurrForce
     Write(Log_Unit,'("MoveAllAtoms: ", A10," V(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Force%Velocity(:,1)


     If ((Mode == RELAX_MD) .AND. (.NOT. Atom_List(i)%Freeze)) then
        Call MoveAtom(Atom_List(i)%Force, &
             AtomType_Info(Atom_List(i)%TypeIndex)%Mass, Atom_List(i)%Friction)
     End If

     Atom_List(i)%Pos = Atom_List(i)%Force%Pos(:,1)  !*** Copy it 
     Atom_List(i)%Frac_Pos = MATMUL(xtal%InverseBasis, Atom_List(i)%Pos)  

     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(new)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos     
     Write(Log_Unit,'("MoveAllAtoms: ", A10," R(xtal) =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Frac_Pos     
  End Do

  Call Flush_(Log_Unit)


 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!******************************************************************************
!
!  RelaxGeometry - Moves the atoms in the direction of the forces on them  
!
!  ForceTol, MaxMove -- determine the scale of the move
!  ConditionNo  -- determines parameters for FitMove
!
!******************************************************************************
Subroutine RelaxGeometry(ForceTol, MaxMove, ConditionNo)
  Real, Intent(IN) :: ForceTol, MaxMove, ConditionNo

  Real :: CurrForce(3), scale,R(3),F(3),energy,Efit
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Real, allocatable :: Edata(:),Rdata(:,:),Fdata(:,:),Rfit(:),Fconstraint(:,:)
  Integer, allocatable :: Map(:)
  Integer :: i, j, k, a,ier,index, nmove,ndim , Nfreedom
  character*4 :: label
  character*9 :: moveoption
  logical :: MatrixMethod,Success


  Write(Log_Unit,*) 'RelaxGeometry: Printing Updated Atom Positions and Forces'
! Read Name_Pos file 
! If there are more than 2 entries ,  call FitMove
  
   Success=.FALSE.

  open(unit=Positions_Unit,file=Trim(Pos_Name),form='formatted', &
          action='read',iostat=ier)
  MatrixMethod=.false.
  index=0
  nmove=0

  If (ier/=0) then
     Write(Error_Unit,*)' Error in RelaxGeometry -- no force/position file',&
             Trim(Pos_Name)
     Stop
  EndIf

    Do 
       read(Positions_Unit,'(a4)',iostat=ier) label
       If (ier/=0) exit
       write(Log_Unit,*) label
       If (label=="STEP") then
          index=index+1
          read(Positions_Unit,*,iostat=ier)  ! energy
          Do i=1,Specific_Atoms
             read(Positions_Unit,'(68x,a9)',iostat=ier) moveoption  ! R
             write(Log_Unit,*) index,i,moveoption
             read(Positions_Unit,*,iostat=ier) ! F
             If (index==1.AND.trim(moveoption)=="") nmove=nmove+1
          Enddo
       EndIf
       write(Log_Unit,*) index,nmove
    Enddo

  Write(Log_Unit,*) ' RelaxGeometry: index,nmove = ', index,nmove
             
  If (index>2) then
     MatrixMethod=.true.
     Nfreedom=3*nmove   ! this will change for "shell" mode
     Allocate(Edata(index),Rdata(Nfreedom,index),Fdata(Nfreedom,index),&
         Map(Nfreedom),Rfit(Nfreedom))
     If (Anchor) Allocate(Fconstraint(3,index))
     Rewind(Positions_Unit)
     Do i=1,index
        Read(Positions_Unit,*)
        Read(Positions_Unit,'(18x,1pe20.10)') energy
         Edata(i)=-energy
           write(Log_Unit,*) 'Energy',Energy
           k=0
           Do  a=1,Specific_Atoms
             read(Positions_Unit,'(14x,3e18.10,a9)') R(:),moveoption
             write(Log_Unit,*) R(:),moveoption
             read(Positions_Unit,'(14x,3e18.10)') F(:)
             write(Log_Unit,*) F(:)
             If (trim(moveoption)=="") then
                Rdata(k+1:k+3,i)=R(:)
                Fdata(k+1:k+3,i)=F(:)
                Map(k+1:k+3)=a
                k=k+3
             EndIf  
             If (Anchor.AND.a==anchorindex) Fconstraint(:,i)=F(:)
           Enddo

     Enddo

     Close(Positions_Unit)

   If (.NOT.Anchor) then
      Call Fitmove(Index,Nfreedom,Edata,Fdata,          &
        Rdata,Efit,Rfit,MaxMove,ConditionNo,Success)
 
      Write (Log_Unit,*) ' Completed FitMove with Success', success
      Write(Log_Unit,'(1p3e15.7)') (Rfit(i),i=1,nfreedom)
      Write(Log_Unit,*)'RelaxGeometry: E0 =', Efit
   Write(Log_Unit,*) 'RelaxGeometry RESULTS -  New                     Dif'
   Do a=1,Specific_atoms
      Inpos(:,a) = Atom_List(a)%Pos
      Outpos(:,a) = Atom_List(a)%Pos
         Do j=1,Nfreedom,3
            If (Map(j)==a) then
               Outpos(:,a) = Rfit(j:j+2)
               Exit
            EndIf
         Enddo
   Write(Log_Unit,'(i5,2x,3f8.3,2x,3f8.3)')a,Outpos(:,a),InPos(:,a)-OutPos(:,a)
   Enddo
  
   EndIf

   If (Anchor) then

      Do i=1,Index   
         Do j=1,Nfreedom,3
            Fdata(j:j+2,i) = Fdata(j:j+2,i) - Fconstraint(1:3,i)
         Enddo
      Enddo
            

      Call Constraintmove(Index,Nfreedom,Edata,Fdata,          &
        Rdata,Efit,Rfit,Fconstraint,MaxMove,ConditionNo,Success)
 
      Write (Log_Unit,*) ' Completed FitMove with Success', success
      Write(Log_Unit,'(1p4e15.7)') (Rfit(i),i=1,nfreedom)
      Write(Log_Unit,*)'RelaxGeometry: E0 =', Efit
   Write(Log_Unit,*) 'RelaxGeometry RESULTS -  New                     Dif'
   Do a=1,Specific_atoms
      Inpos(:,a) = Atom_List(a)%Pos
      Outpos(:,a) = Atom_List(a)%Pos
         Do j=1,Nfreedom,3
            If (Map(j)==a) then
               Outpos(:,a) = Rfit(j:j+2)
               Exit
            EndIf
         Enddo
   Write(Log_Unit,'(i5,2x,3f8.3,2x,3f8.3)')a,Outpos(:,a),InPos(:,a)-OutPos(:,a)
   Enddo
  
   EndIf


   If (Success) then
       Do a=1,Specific_atoms
          Atom_List(a)%Force%Pos(:,1) = Atom_List(a)%Pos
       Enddo

       Do i=1,nfreedom,3  ! This will change with shell option
          a=Map(i)
          Atom_List(a)%Force%Pos(:,1)= Rfit(i:i+2)
       Enddo

       Write(Log_Unit,*) 'RelaxGeometry --   New                     Dif'
       Do a=1,Specific_Atoms
          Write(Log_Unit,'(i5,2x,3f8.3,2x,3f8.3)') a,&
              Atom_List(a)%Force%Pos(:,1),&
              Atom_List(a)%Force%Pos(:,1)-Atom_List(a)%Pos

              Atom_List(a)%Pos = Atom_List(a)%Force%Pos(:,1)  !*** Copy it 
              Atom_List(a)%Frac_Pos=MATMUL(xtal%InverseBasis,Atom_List(a)%Pos)  

       Enddo

    EndIf

  Else  

     Close(Positions_Unit)
    
  EndIf

  If (.NOT.Success) Call ForceMove(ForceTol,MaxMove)
  
  Call Flush_(Log_Unit)

 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! ConstraintMove
!
!   Estimates equilibrium values of positions given          
!     Niter values of E,  forces,  positions , and constraint forces
!
!   Edata is an array of the total energies (or negative of cohesive E)
!      for each iteration
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Subroutine ConstraintMove(Niter,Nfreedom,Edata,Fdata,Rdata,Efit,Rfit,&
           Fconstraint,Maxmove,ConditionNo,Success)

  Integer, Intent(IN) :: Niter, Nfreedom
  Real,    Intent(IN) :: Edata(Niter),Rdata(Nfreedom,Niter),Maxmove,ConditionNo
  Real,    Intent(IN) :: Fdata(Nfreedom,Niter),Fconstraint(3,Niter)
  Real,    Intent(OUT) :: Efit,Rfit(Nfreedom)
  Logical, Intent(OUT) :: Success

  Real, allocatable  :: A(:,:),U(:,:),VT(:,:),S(:),Work(:),Rref(:),B(:)
  Integer, allocatable :: IWork(:),Index(:)
  Integer :: N, M, Lwork, Liwork, Info, i,j,k,icount
  Real :: x

  Success = .false.
  Efit=0;Rfit=0

  If (Niter.lt.2) Return

  M=2*Niter
  N=Nfreedom+1

  i = MAX(N,M)   ; icount=MIN(M,N)
  Lwork=8*i*i+12*i
  Liwork=8*i

  Allocate(A(M,N),S(N),U(M,M),VT(N,N),Work(Lwork),IWork(Liwork),&
           Rref(Nfreedom),B(M),Index(Niter),stat=info)
    If (info /= 0) then
       write(Error_Unit,*) 'Error in ConstraintMove allocation ', info
       call flush_(Error_Unit)
       stop
    Endif

  Call Insertion_Sort(Edata(1:Niter),Index,.TRUE.)  !Order in increasing E

  ! Use lowest energy iteration as reference

  Rref(1:Nfreedom)=Rdata(1:Nfreedom,Index(1))

  A(1:Niter,1)=1
  A(Niter+1:M,1)=0

  Do i=1,Niter
       j=Index(i)
        A(i,2:N) = Fdata(1:Nfreedom,j)/2
        B(i)=Edata(j)+&
                 DOT_PRODUCT(Fdata(1:Nfreedom,j),&
                   (Rdata(1:Nfreedom,j)-Rref(1:Nfreedom)))/2
        
        B(Niter+i)=0
        Do k=1,Nfreedom,3
           A(Niter+i,k+1:k+3) = Fconstraint(1:3,j)
           B(Niter+i) = B(Niter+i) + & 
              DOT_PRODUCT(Fconstraint(1:3,j),(Rdata(k:k+2,j)-Rref(k:k+2)))
        Enddo

  EndDo

  Call DGESDD('A',M,N,A,M,S,U,M,VT,N,Work,Lwork,Iwork,info)
     Write(Log_Unit,*) 'completed DGESDD with info = ',info
     Write(Log_Unit,*) ' singular values '
     Write(Log_Unit,'(1p5e15.7)') (S(i),i=1,icount)
     If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove RDGESDD ', info
       call flush_(Error_Unit)
       stop
     Endif
  
  Work=0
  x=ABS(S(1))/ConditionNo
  Do i=1,icount
     If (ABS(S(i)).gt.x) then
        Work(1:N)=Work(1:N)+VT(i,1:N)*DOT_PRODUCT(U(1:M,i),B)/S(i)
     EndIf
  Enddo

  Efit=Work(1)
  Rfit(1:Nfreedom)=Rref(1:Nfreedom)+Work(2:N)      

   write(Log_Unit,*) ' Efit = ',Efit

   write(Log_Unit,*) ' Rfit = ',Rfit

 ! Check MaxMove:
      x = MAXVAL(ABS(Work(2:N)))

    Write(Log_Unit,*) 'error measure = ',x

   if (x <= MaxMove) Success=.true.
   Deallocate(A,S,U,VT,Work,IWork,Rref,B,Index)
 
   Return

End subroutine ConstraintMove

!******************************************************************************
!
!  ForceMove_old - Moves the atoms in the direction of the forces on them  
!
!  ForceTol, MaxMove -- determine the scale of the move
!
!******************************************************************************

Subroutine ForceMove_old(ForceTol, MaxMove)
  Real, Intent(IN) :: ForceTol, MaxMove

  Real :: CurrForce(3), scale
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Integer :: i, j

  Write(Log_Unit,*) 'ForceMove: Moving atoms in direction of forces'

   scale=0
  Do i=1, Specific_Atoms
  scale=scale+DOT_PRODUCT(Atom_List(i)%Force%CurrForce,&
        Atom_List(i)%Force%CurrForce)
  EndDo

  If (scale > 0) scale=SQRT(scale)

      scale= Max(scale/Specific_Atoms,ForceTol)
        scale= MaxMove/scale


    Write(Log_unit,*) 'ForceMove -- move scale = ', scale
  
  Do i=1, Specific_Atoms 
           !** Calculate the new position *
     Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Pos  &
                                +  Atom_List(i)%Force%CurrForce*scale
     Write(Log_Unit,'("ForceMove: ", A10," R(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos
     Write(Log_Unit,'("ForceMove: ", A10," F(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Force%CurrForce

     Atom_List(i)%Pos = Atom_List(i)%Force%Pos(:,1)  !*** Copy it 
     Atom_List(i)%Frac_Pos = MATMUL(xtal%InverseBasis, Atom_List(i)%Pos)  

     Write(Log_Unit,'("ForceMove: ", A10," R(new)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos     
     Write(Log_Unit,'("ForceMove: ", A10," R(frac) =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Frac_Pos     
  End Do

  Call Flush_(Log_Unit)


 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!******************************************************************************
!
!  ForceMove - Moves the atoms in the direction of the forces on them  
!
!  ForceTol, MaxMove -- determine the scale of the move
!
!******************************************************************************

Subroutine ForceMove(ForceTol, MaxMove)
  Real, Intent(IN) :: ForceTol, MaxMove

  Real :: CurrForce(3), scale ,x(3*specific_atoms), u
  Real :: Inpos(3,specific_atoms),Outpos(3,specific_atoms)
  Integer :: i, j

  Write(Log_Unit,*) 'ForceMove: Moving atoms in direction of forces'

   scale=0
  Do i=1, Specific_Atoms
   CurrForce=Atom_List(i)%Force%CurrForce
   If (Anchor) CurrForce=CurrForce-Atom_List(Anchorindex)%Force%CurrForce
   scale=MAX(scale,ABS(CurrForce(1)),ABS(CurrForce(2)),ABS(CurrForce(3)))
  EndDo

  scale= Max(scale,MaxMove)
  scale= MaxMove/scale

    Write(Log_unit,*) 'ForceMove -- move scale = ', scale
  
  CurrForce = 0
  If (Anchor) CurrForce=Atom_List(Anchorindex)%Force%CurrForce
  
  Do i=1, Specific_Atoms 
           !** Calculate the new position *
     Atom_List(i)%Force%Pos(:,1) = Atom_List(i)%Pos  &
           +  (Atom_List(i)%Force%CurrForce-CurrForce)*scale
     Write(Log_Unit,'("ForceMove: ", A10," R(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos
     Write(Log_Unit,'("ForceMove: ", A10," F(old)  =",1p3e13.5)') &
             Atom_List(i)%Name, (Atom_List(i)%Force%CurrForce-CurrForce)

     Atom_List(i)%Pos = Atom_List(i)%Force%Pos(:,1)  !*** Copy it 
     Atom_List(i)%Frac_Pos = MATMUL(xtal%InverseBasis, Atom_List(i)%Pos)  

     Write(Log_Unit,'("ForceMove: ", A10," R(new)  =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Pos     
     Write(Log_Unit,'("ForceMove: ", A10," R(frac) =",1p3e13.5)') &
             Atom_List(i)%Name, Atom_List(i)%Frac_Pos     
  End Do

  Call Flush_(Log_Unit)


 If (Xtal%Rot_Size > 1) then
    Do i=1,Specific_Atoms
       Inpos(:,i)=Atom_List(i)%Frac_Pos
    Enddo
    Call Symmetrize_Pos(Inpos,Outpos)
    Do i=1,Specific_Atoms
       Atom_List(i)%Frac_Pos=Outpos(:,i)
       Atom_List(i)%Pos=MATMUL(Xtal%Basis,Atom_List(i)%Frac_Pos)
    Enddo
 EndIf

  Return
End Subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Fitmove
!
!   Estimates equilibrium values of generalized positions given          
!     Niter values of E, generalized forces, and generalized positions
!
!   Edata is an array of the total energies (or negative of cohesive E)
!      for each iteration
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Subroutine Fitmove(Niter,Nfreedom,Edata,Fdata,Rdata,Efit,Rfit,&
           Maxmove,ConditionNo,Success)

  Integer, Intent(IN) :: Niter, Nfreedom
  Real,    Intent(IN) :: Edata(Niter),Rdata(Nfreedom,Niter),Maxmove,ConditionNo
  Real,    Intent(IN) :: Fdata(Nfreedom,Niter)
  Real,    Intent(OUT) :: Efit,Rfit(Nfreedom)
  Logical, Intent(OUT) :: Success

  Real, allocatable  :: A(:,:),U(:,:),VT(:,:),S(:),Work(:),Rref(:),B(:)
  Integer, allocatable :: IWork(:),Index(:)
  Integer :: N, M, Lwork, Liwork, Info, i,j,k,icount
  Real :: x

  Success = .false.
  Efit=0;Rfit=0

  If (Niter.lt.2) Return

  M=Niter
  N=Nfreedom+1

  i = MAX(N,M)   ; icount=MIN(M,N)
  Lwork=8*i*i+12*i
  Liwork=8*i

  Allocate(A(M,N),S(N),U(M,M),VT(N,N),Work(Lwork),IWork(Liwork),&
           Rref(Nfreedom),B(M),Index(Niter),stat=info)
    If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove allocation ', info
       call flush_(Error_Unit)
       stop
    Endif

  Call Insertion_Sort(Edata(1:Niter),Index,.TRUE.)  !Order in increasing E

  ! Use lowest energy iteration as reference

  Rref(1:Nfreedom)=Rdata(1:Nfreedom,Index(1))

  A(1:M,1)=1

  Do i=1,M
       j=Index(i)
        A(i,2:N) = Fdata(1:Nfreedom,j)/2
        B(i)=Edata(j)+&
                 DOT_PRODUCT(Fdata(1:Nfreedom,j),&
                   (Rdata(1:Nfreedom,j)-Rref(1:Nfreedom)))/2
  EndDo

  Call DGESDD('A',M,N,A,M,S,U,M,VT,N,Work,Lwork,Iwork,info)
     Write(Log_Unit,*) 'completed DGESDD with info = ',info
     Write(Log_Unit,'(1p5e15.7)') ' singular values ', (S(i),i=1,icount)
     If (info /= 0) then
       write(Error_Unit,*) 'Error in Fitmove RDGESDD ', info
       call flush_(Error_Unit)
       stop
     Endif
  
  Work=0
  x=ABS(S(1))/ConditionNo
  Do i=1,icount
     If (ABS(S(i)).gt.x) then
        Work(1:N)=Work(1:N)+VT(i,1:N)*DOT_PRODUCT(U(1:M,i),B)/S(i)
     EndIf
  Enddo

  Efit=Work(1)
  Rfit(1:Nfreedom)=Rref(1:Nfreedom)+Work(2:N)      

   write(Log_Unit,*) ' Efit = ',Efit

   write(Log_Unit,*) ' Rfit = ',Rfit

 ! Check MaxMove:
      x=-100
      Do i=1,Nfreedom
         x=Max(x,ABS(Rref(i)-Rfit(i)))
      Enddo

    Write(Log_Unit,*) 'error measure = ',x

   if (x <= MaxMove) Success=.true.
   Deallocate(A,S,U,VT,Work,IWork,Rref,B,Index)
 
   Return

End subroutine FitMove



!******************************************************************************
!
!  PrepPrevious - Prepares the previous set of Psi's for use with the
!      new positions
!
!******************************************************************************

Subroutine PrepPrevious
  Integer :: i, Psi_toUse(Mem_MapSize), VecSize,  j, Flag ,k
  Complex, Pointer  :: Vec(:,:), Work(:)
  Real     :: Eigen(TotalPsi),  Curr(2), mag, Energy
  Integer  :: NewPsi_Map(Mem_MapSize), NewSize
  Integer  :: LUT(2,Mem_MapSize)
  Type (Mem_Handle), Pointer :: Psi



  Call GetBuffer( Work)

  DoLCAO = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  
  curr = (/V_NewMix, Dij_NewMix/)
  V_NewMix = 1;  Dij_NewMix = 1;

  !*** First Normalize the Psi's to Qt+Qh=0 ***
  
  NewPsi_Map = MH_Skip

  Do  k=1, NumKpnts
     Psi_toUse = MH_Skip
     Do i=1,Mem_MapSize
       If ((PsiInfo(i)%Kpnt == k) .AND. &
               (PsiInfo(i)%Available == Mem_Used)) then
        Psi_toUse(i)=MH_toProcess
        NewPsi_Map(i)=MH_toProcess
       EndIf
     EndDo

     Call Phase_Generic( Psi_toUse,  k)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0)
        j = Psi%Index
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
        PsiInfo(j)%PDOT_Stored = .FALSE.

        Call Opsi( Psi%Ptr, Work, &
             PsiInfo(j)%PDOT, i, PsiInfo(j)%PDOT_Stored)
              
        mag = Basis_DotProd( Psi%Ptr, Work, .FALSE.)
        mag = SQRT(mag)
!Write(Log_Unit,*) 'PrepPrev: Kpnt=',i, ' * DI=',Psi%Index, ' * DOT=',mag
        Psi%Ptr = Psi%Ptr / mag
        PsiInfo(Psi%Index)%PDOT = PsiInfo(Psi%Index)%PDOT/mag

        Call SavePsi_toBuffer(j,Psi%Ptr)
        Call GetNextPsi(Flag, Psi)
     End Do
  End Do

! Clean up the remaining stored wavefunctions

  Do j=1,Mem_MapSize
     If(NewPsi_Map(j) == MH_Skip) then
       PsiInfo(j)%Kpnt=0
       PsiInfo(j)%DoSave=0
       PsiInfo(j)%Available=Mem_Available
       i=PsiInfo(j)%MemBufIndex
       If(i>0) Globalmap%Psi_Handle(i)%Index=0
       PsiInfo(j)%MemBufIndex=0
     EndIf
  Enddo

  Call CalcMeritFunction(  Energy)

  Do i=1, NumKpnts
     Psi_toUSe = MH_Skip
     Where (PsiInfo(:)%Kpnt == i) Psi_toUse = MH_toProcess
     VecSize = SUM(Psi_toUse)
     Allocate(Vec(VecSize, VecSize))


     Call PerformGRRP_ALL( Psi_toUse, VecSize, &
       Vec, Eigen, LUT, NewSize, i)
     Call MakeNewPsi( VecSize, NewSize, Vec, Eigen, LUT, i)

     DeAllocate(Vec)
  End Do
Write(Log_Unit,*) 'PrepPrevious   : After Orthog'

!ReCalculate Occupancies
  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_ALL

  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  Call CalcMeritFunction(  Energy)

  V_NewMix = curr(1);  Dij_NewMix = curr(2); 
  Call CalcEigenvalues

  Call FreeBuffer(Work)

  Return
End Subroutine

!******************************************************************************
!
!  PrepPrevious_band bandstructure_mode  - Prepares the previous set of Psi's
!      next iteration
!
!******************************************************************************

Subroutine PrepPrevious_band(NumB)
  Integer, Intent(IN) :: NumB
  Integer :: i, Psi_toUse(Mem_MapSize), VecSize,  j, Flag,n,maxindex
  Complex, Pointer  :: Vec(:,:), Work(:)
  Real     :: Eigen(TotalPsi), Curr(2), mag, Energy
  Integer  :: NewSize
  Integer  :: LUT(2,Mem_MapSize)
  Type (Mem_Handle), Pointer :: Psi

  MinBands=NumB+4
  TotalPsi=MinBands+8
  Numbands=NumB

  Call GetBuffer( Work)

  DoLCAO = 0
  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  

! Find set of bands from previous run with energies less than Eigen_Max

  Psi_toUse = MH_Skip

  n=0; maxindex=Maxval(PsiInfo(:)%Kpnt)

  Kloop: Do i=1,maxindex
    Do  j=1,Mem_MapSize
     If (PsiInfo(j)%Available == Mem_Used) then
        If ((PsiInfo(j)%Kpnt == i) .AND. (PsiInfo(j)%Energy <= Eigen_max)) then
           Psi_toUse(j) = MH_toProcess
           n=n+1
           If (n > NumB) Exit Kloop
        EndIf
     EndIf
    ENddo
  End Do  Kloop
     
  If (n < NumB) then
    DoLcao=1
    return
  Endif

  i=1
  !  Clean up the remaining stored wavefunctions
  Do j = 1,Mem_MapSize
     If (Psi_toUse(j)== MH_toProcess) then
        PsiInfo(j)%Kpnt=i
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
     Else
        PsiInfo(j)%Kpnt=0
        PsiInfo(j)%DoSave=0
        PsiInfo(j)%Available=Mem_Available
        n=PsiInfo(j)%MemBufIndex
        If (n > 0) then
           Globalmap%Psi_Handle(n)%Index=0
        EndIf
        PsiInfo(j)%MemBufIndex=0
     EndIf
  Enddo

     Call Phase_Generic( Psi_toUse,  i)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0)
        j = Psi%Index
        PsiInfo(j)%DoSave=1
        PsiInfo(j)%Available=Mem_Used
        PsiInfo(j)%PDOT_Stored = .FALSE.

        Call Opsi( Psi%Ptr, Work, &
             PsiInfo(j)%PDOT, i, PsiInfo(j)%PDOT_Stored)
              
        mag = Basis_DotProd( Psi%Ptr, Work, .FALSE.)
        mag = SQRT(mag)
Write(Log_Unit,*) 'PrepPrev: Kpnt=',i, ' * DI=',Psi%Index, ' * DOT=',mag
Call flush_(log_unit)
        Psi%Ptr = Psi%Ptr / mag
        PsiInfo(Psi%Index)%PDOT = PsiInfo(Psi%Index)%PDOT/mag

        Call SavePsi_toBuffer(j,Psi%Ptr)
        Call GetNextPsi(Flag, Psi)
     End Do

  Call CalcMeritFunction(  Energy)

  Do i=1, NumKpnts
     Psi_toUSe = MH_Skip
     Where (PsiInfo(:)%Kpnt == i) Psi_toUse = MH_toProcess
     VecSize = SUM(Psi_toUse)
     Allocate(Vec(VecSize, VecSize))

     Call PerformGRRP_ALL( Psi_toUse, VecSize, &
       Vec, Eigen, LUT, NewSize, i)
     Call MakeNewPsi( VecSize, NewSize, Vec, Eigen, LUT, i)

     DeAllocate(Vec)
  End Do
Write(Log_Unit,*) 'PrepPrevious_band   : After Orthog'

!ReCalculate Occupancies
  Call MakeBandMap
  Call CalcNewOccAndFermi
  Call ReCluster_ALL

  PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear the PDOT's
  PsiInfo(:)%KE_Stored = .FALSE.    !** and KE's
  Call CalcMeritFunction(  Energy)

  V_NewMix = curr(1);  Dij_NewMix = curr(2); 
  Call CalcEigenvalues

  Call FreeBuffer(Work)

  Return
End Subroutine

!******************************************************************************
!
! RelaxAtoms - Relaxes the Atoms to equilibrium or until the max number of
!       allowed steps are made.
!
!  Mode      - Relaxation mode 0=MD, 1=Geometry
!  MaxStep   - Max number of allowed steps
!  ChargeBase - Ouput base filename containing the Psi's and V for each 
!               step.  If ChargeBASE="NULL" then nothing is stored   
!  MaxSCF     - Max number of SCF steps to perform/atom iteration
!  ChargeTol  - Cohesive Energy tolerance for RelaxElectrons
!  ForceTol   - Force Magnitude tolerance to determine equilibrium
!  MaxMove    - Maximum move distance allowed for Geometry optimization
!  ConditionNo -- Parameter for Fitmove in Geometry optimization
!
!******************************************************************************

Subroutine RelaxAtoms(Mode,MaxStep,ChargeBase,MaxSCF,ChargeTol,ForceTol,&
                             MaxMove,ConditionNo)
  Integer,       Intent(IN) :: Mode
  Integer,       Intent(IN) :: MaxStep
  Character*(*), Intent(IN) :: ChargeBase
  Integer,       Intent(IN) :: MaxSCF
  Real,  Intent(IN) :: ChargeTol
  Real,  Intent(IN) :: ForceTol
  Real,  Intent(IN) :: MaxMove , ConditionNo 

  Character*100 :: FileName,message
  Logical       :: Converged, EConverged
  Integer       :: Fstep, i
  Real          :: TotalE, KE, curr(2)
  

  Converged = .FALSE.
  Fstep = 1

  Call RelaxElectrons(MaxSCF, ChargeTol, EConverged) 
  If (Trim(ChargeBase) /= "NULL") then
     Write(Filename ,*) FStep
     FileName = Trim(ChargeBase) // Trim(ADJUSTL(Filename)) // '.rsgrid'
     Call StoreSolution_Binary( FileName)
  End If
     
  If (EConverged) message='SCF step converged'
  If (.not.EConverged) message='SCF step not converged'
  Call CalcForces
  Call Store_ForcesAndPositions(Trim(Force_Name),message)
  If (EConverged) Call Position_Archive
  Call CheckForces(ForceTol,Converged)

  !If (Mode == RELAX_GEOMETRY) Call MakeGeometryMixing

  Do While ((.NOT. Converged) .AND. (Fstep < MaxStep))
     Fstep = Fstep + 1

     !*** Calculate the KE of the nuclei **
     KE = 0
     Do i=1, Specific_Atoms
        KE = KE + 0.5*AtomType_Info(Atom_List(i)%TypeIndex)%Mass* &
             SUM(Atom_list(i)%Force%Velocity(:,1)**2)
     End Do

     TotalE = CohesiveEnergy - KE
     Write(Log_Unit,*) 'RelaxAtoms: Total energy:',TotalE, ' * Cohesive=', &
          CohesiveEnergy, ' * KE=',KE

     If (Mode == RELAX_MD) Call MoveAllAtoms(Mode)
! For relax_geometry mode, do not move if SCF step has not converged
     If (Mode == RELAX_GEOMETRY .AND. EConverged) then
               Call RelaxGeometry(ForceTol,MaxMove,ConditionNo)
               Write(Output_Unit,*)
               Write(Output_Unit,*) 'Atoms moved to new positions'
               Write(Output_Unit,*)
               Call Flush_(Output_Unit)
     EndIf
     If (Mode == RELAX_GEOMETRY .AND. .not.EConverged) then
               Write(Output_Unit,*)
               Write(Output_Unit,*) 'Atoms left in same positions'
               Write(Output_Unit,*)
               Call Flush_(Output_Unit)
     EndIf
     If(.not.bandstructure_mode)Call StoreSolution_Binary('paw.wfn')

     !** First reset the mixing to NULL ***
     Call Anderson_ResetMix(SCFvalues%AC)
     Do i=1, Specific_Atoms
        Call Anderson_ResetMix(Atom_List(i)%Dij_AC)
     End Do
     
     Call UpdateSystem

     If (MD_Guess == MD_PREVIOUS) then
        Call PrepPrevious
!Call LoadGuess('paw.wfn')
     else  
        DoLCAO = Initial_guess
        !*** Make sure ALL wave functions are free ***
        PsiInfo(:)%Kpnt = 0
        PsiInfo(:)%DoSave = 0
        PsiInfo(:)%Available = Mem_Available
        PsiInfo(:)%PDOT_Stored = .FALSE.
        PsiInfo(:)%KE_Stored = .FALSE.    
     End If
     
     
     Call RelaxElectrons(MaxSCF, ChargeTol, EConverged) 
     If (Trim(ChargeBase) /= "NULL") then
        Write(Filename ,*) FStep
        FileName = Trim(ChargeBase) // Trim(ADJUSTL(Filename)) // '.wfn'
        Call StoreSolution_Binary( FileName)
     End If

     If (EConverged) message='SCF step converged'
     If (.not.EConverged) message='SCF step not converged'
     Call CalcForces
     Call Store_ForcesAndPositions(Trim(Force_Name),message)
     If (EConverged) Call Position_Archive
     Call CheckForces(ForceTol,Converged)
  End Do
  
  Return
End Subroutine


!******************************************************************************
!
! RelaxTime - Loads a time-dependent run fron the input file
!
!   NumSteps - Number of steps to take
!
!   The calling format is :
!
!       SET_NAME TOTAL_CURRENT_NAME 'current_filename'  # optional
!       SET_NAME ELECTRIC_FIELD_NAME 'e_field_name'     # optional
!       TIME_STEP  dt                                   # optional
!
!       RELAX TIME Num_steps    # This is the actual command
!
!******************************************************************************

Subroutine RelaxTime(NumSteps)
  Integer, Intent(IN) :: NumSteps

  Integer :: n, Psi_toUSe(Mem_MapSize), i, Kpnt,Mode
  Type (Mem_Handle), Pointer :: Psi
  Complex, Pointer :: NewPsi(:)
  Real :: backup(2), time, Current(3), Elec(3)
  Logical :: AtomsMoved

  Write(*,*) "RelaxTime --  Mode has not been defined"

  AtomsMoved = .FALSE.
  Do i=1, Specific_Atoms
     If (.NOT. Atom_list(i)%Freeze) AtomsMoved = .TRUE.
  End Do

  Write(Log_Unit,*) 'RelaxTime: Initial settings for time-dependent run'
  Write(Log_Unit,*) '   Total Steps:',Numsteps, ' * DT=',TimeStep, ' fs'
  Write(Log_Unit,*) '   Total simulation time :',NumSteps*TimeStep, ' fs'
  Write(log_Unit,*) ' '
  Write(Log_Unit,*) '   Force Output:', TRIM(Force_Name)
  Write(Log_Unit,*) '   Total Current Name:', TRIM(TotalCurrent_Name)
  Write(Log_Unit,*) '   Electric Field Name:', TRIM(E_Name)
  Write(Log_Unit,*) '   ** NOTE: Appending ALL results to the above files! **'
  Write(log_Unit,*) ' '

  open(unit=TotalCurrent_Unit,file=Trim(totalCurrent_Name), RECL=1024, &
       position='append')
  Write(Totalcurrent_Unit,*)
  Call Printdate(Totalcurrent_Unit,'#  Current results generated on date ')
  Write(TotalCurrent_Unit, *) '# 1.Time  2.Jx   3.Jy  4.Jz'  

  open(unit=E_Unit,file=Trim(E_Name), RECL=1024, &
       position='append')
  Write(E_Unit,*)
  Call Printdate(E_Unit,'#  E-field results generated on date ')
  Write(E_Unit, *) '# 1.Time  2.Ex   3.Ey  4.Ez  5.Ax  6.Ay  7.Az'  

  backup = (/V_newMix, Dij_NewMix/)
  V_newMix=1
  Dij_NewMix=1

  time = 0
  Call CalcVectorPotential(time, Elec, Vector_Pot)
  Call CalcTotalCurrent(Current)  
Write(Log_Unit,*) 'RelaxTime: Elec=',Elec, ' * VecPot=',Vector_Pot
Write(Log_Unit,*) 'RelaxTime: Atomic_Mode=',Atomic_Mode
Call CalcEigenvalues

  Atomic_Mode = SIM_TIME
Write(Log_Unit,*) 'RelaxTime: SIM_TIME * AM=',Atomic_Mode
Call CalcEigenvalues

  Do n=1, NumSteps
     Write(Log_Unit,*) 'RelaxTime:  Iter=',n
     Call PrintDate(Log_Unit,'RelaxTime: Start')

     !** Update all Loaded bands **
     Call GetBuffer( NewPsi)
     Do Kpnt = 1, NumKpnts
        Psi_toUse = MH_Skip
        Where (PsiInfo(:)%Kpnt == Kpnt) Psi_toUse = MH_toProcess

        Call Phase_Generic( Psi_toUse,  kpnt)
        
        Call GetNextPsi(i, Psi)
        Do While (i /= 0)
           Call TIME_PropagatePsi(Psi%Index, Psi%Ptr, NewPsi, Kpnt)        
           Call SavePsi_toBuffer( Psi%Index, NewPsi)

           Call GetNextPsi(i, Psi)
        End Do
     End Do
     Call FreeBuffer(NewPsi)
     
     !** Move the atoms **     
     If (AtomsMoved) then
        Call CalcForces
        Call Store_ForcesAndPositions(Force_Name,'Called from RelaxTime')
        Call MoveAllAtoms(Mode)
        Call UpdateSystem
     End If

     time = TimeStep*(n-1)
     Call CalcVectorPotential(time, Elec, Vector_Pot)
     Call CalcTotalCurrent(Current)

     Write(TotalCurrent_Unit,*) time, Current
     Write(E_Unit,*) time, Elec, Vector_Pot
     
     !** Update the hamiltonian **
     Call CalcHam(CohesiveEnergy)
     PsiInfo(:)%PDOT_Stored = .FALSE.
     Call CalcEigenvalues
  End Do

  V_newMix = Backup(1)
  Dij_NewMix = Backup(2)

  Close(TotalCurrent_Unit)
  Close(E_Unit)

  Write(Log_Unit,*) 'RelaxTime: Completed Run!'

  Atomic_Mode = SIM_PAW

  Return
End Subroutine


!******************************************************************************
!
! LDRelax - Processes the RELAX statment which has the form
!
!  RELAX CHARGE  MaxIter Tol  ---- Relax the electrons
!  RELAX [ATOMS|GEOMETRY] MaxSteps ChargeBase MaxIter ChargeTol ForceTol &
!                        MaxMove ConditionNo
!  RELAX TIME Num_Steps 
!
!******************************************************************************

Subroutine LdRelax(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, MaxSteps, mode
  Real  :: Tol, ForceTol, Bo, MaxMove, ConditionNo
  Logical :: EConverged

  Call GetNextWord(WC, Token, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdRelax: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdRelax:")
     STOP
  End If

  Call UpperCase(token)
  If (trim(Token) == 'CHARGE') then
     Call GetNumbers(WC, MaxIter)
     Call GetNumbers(WC, Tol)

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Relaxing Charge:  MaxIter=",MaxIter, &
             ' * Tolerance=',Tol
        Write(LOG_Unit,*) 'ldRelax: Starting Cohesive Energy :', CohesiveEnergy
     End IF

     If (PAW_FirstTime) then
        PAW_FirstTime = .FALSE.
     End if

     Call RelaxElectrons(MaxIter, Tol, EConverged) 

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Completed Charge relaxation."
     End IF
  else if ((Trim(Token) == 'ATOMS') .OR. (Trim(Token) == 'GEOMETRY')) then
     If (trim(Token) == 'ATOMS') then
        Mode = RELAX_MD
     else
        Mode = RELAX_GEOMETRY
     End If

     Call GetNumbers(WC, MaxSteps)
!*     Call GetNextWord(WC, Filename, tlen)
     Call GetNextWord(WC, ChargeBase, tlen)
     Call GetNumbers(WC, MaxIter)
     Call GetNumbers(WC, Tol)
     Call GetNumbers(WC, ForceTol)
     Call GetNumbers(WC, MaxMove)
     Call GetNumbers(WC, ConditionNo)

     if (Print_Level == PRINT_VERBOSE) then
        write(LOG_Unit, *) "ldRelax: Relaxing Atoms:  MaxSteps=", MaxSteps, &
             ' * MaxChargeITer=',MaxIter, ' * Force Tolerance=', ForceTol,  &
             ' * Charge Tolerance=',Tol, ' * Max Move=', MaxMove, &
             ' * Condition Number = ', ConditionNo
        Write(LOG_Unit,*) 'ldRelax: Starting Cohesive Energy :', CohesiveEnergy
     End IF


     Call RelaxAtoms(mode,MaxSteps, TRIM(ChargeBase), MaxIter, Tol,ForceTol,&
         MaxMove,ConditionNo)
  else if (TRIM(Token) == "TIME") then
     Call GetNumber(WC, MaxSteps)
     Call RelaxTime(MaxSteps)
  else
     Write(Error_Unit,*) 'LdRelax: Unknown RELAX option: ',TRim(Token)
  End If

  Return
End Subroutine

!******************************************************************************
!
!  LdBandStruct - Loads the Band structure information from disk and
!        processes it.
!
!******************************************************************************

Subroutine LdBandStruct(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, i,j,k, Kindex, NumK, NumB
  Real  :: Tol, Kpnt(3), FracKpnt(3),ENK  
  Type (Word_Context) :: Kpnt_WC
  Logical :: EConverged

  bandstructure_mode=.true.
  ReCalculate_Occupancies=.false.

  Call GetNextWord(WC, Filename, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdBandStruct: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdBandStruct:")
     STOP
  End If

  Band_Name=Trim(FileName)//'.band'

  open(unit=Band_Out,file=Band_name,form='formatted',position='append')

  Call GetNumbers(WC, MaxIter)
  Call GetNumbers(WC, Tol)
  
  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) "ldBandStruct: Relaxing Charge:  MaxIter=",MaxIter, &
          '* Tolerance=',Tol, ' * K-point File:', TRIM(FileName)
  End IF

  If (.not.HamLoaded) then
     write(Error_Unit,*) ' Error in Band Structure Calculation -- ', &
       'Must first calculate or load Hamiltonian'
     stop
  EndIf

  Call InitWord(Kpnt_WC, BAND_IN, Filename, DELIMS, COMMENT, LIT_CHAR, &
       1, INCLUDE_STR)


  !*** Create the New K-points storage space ***
  NumKpnts = 1
  BZ%TotalUniq = NumKpnts
  BZ%TotalKpnts = NumKpnts

write(log_unit,*) 'ldbandstruct -- before allocations';call flush_(Log_Unit)
  If (Associated(bz%ku)) Nullify(Bz%ku) 
  If (Associated(bz%keu)) Nullify(Bz%keu) 
  If (Associated(bz%wtuniq)) Nullify(Bz%wtuniq) 
  If (Associated(bz%weight)) Nullify(Bz%weight)
  If (Associated(xtal%wt)) Nullify(xtal%wt) 
  If (Associated(occupancy)) Nullify(occupancy) 
  If (Associated(User_Kpoints)) Nullify(User_Kpoints) 

  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(NumBands,1), &
       BZ%WtUniq(NumBands, 1), BZ%Weight(NumBands, 1), &
       Occupancy(NumBands, NumKpnts),  User_Kpoints(3,1), stat=i)

write(log_unit,*) 'ldbandstruct -- after allocations',i;call flush_(Log_Unit)
  If (i /= 0 ) then
      write(Error_Unit,*) 'Error in LdBandStructure --' ,&    
        ' Cannot reallocate kpoint arrays ',i
      stop
  EndIf
  BZ%Kp => BZ%Ku;  
  BZ%Weight(:,1) = 1;  !***BZ%Weight(:,2) = 0;
  BZ%WtUniq = BZ%Weight

  BZ%Kp = 0
  User_Kpoints = 0


  !*** Now Get the maximum number of bands for processing
  Call GetNumbers(Kpnt_WC, NumB)      ! Maximum number of bands to calculate

  Write(Log_Unit,*) 'LdBandStruct: Processing ',NumK, ' k-points',NumB,' bands'
  call flush_(Log_Unit)

  Call PrepPrevious_band(NumB)

  !*** Finally process the K's ****
  i=1
  Kpnt = 0
  Kindex=1
  Do 
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) exit

     Write(Log_Unit,*) 'LdBandStruct: Loop=',i, ' * FracKpnt=',FracKpnt
     call flush_(Log_Unit)

     Kpnt(:) = MATMUL(xtal%Recip, FracKpnt)
     User_Kpoints(:,Kindex) = FracKpnt
     Call UpdateKpoint(Kpnt)
     
     Call RelaxElectrons(MaxIter, Tol, EConverged) 

     !*** Write the info to the OUTPUT file ***
     Write(Output_Unit,*) 'LdBandStruct: Loop:',i, ' * FracKpnt=',FracKpnt

     k = 0
     Do j=1, Mem_MapSize
        If (PsiInfo(j)%Kpnt>0) then
           k = k + 1
           ENK=PsiInfo(j)%Energy
           Write(Output_Unit,*) k, j, ENK, PsiInfo(j)%Error
           If (ENK <= Eigen_Max) & 
           Write(Band_Out,'(1p3e15.7,2x,1pe15.7)')Frackpnt,ENK
        End If
     End Do

     Call Flush_(Band_Out)
 
     i=i+1

  End Do

   Write(Log_Unit,*) 'Ending band structure calculation after processing ', &
     i-1,' k-points'
   Call Flush_(Log_Unit)
    

  Call CloseWord(Kpnt_WC)
  bandstructure_mode=.false.

  Return
End Subroutine

!******************************************************************************
!
!  LdSurfBands - Loads the Surface bandstructure info and 
!                  Calculates surface bands
!
!******************************************************************************

Subroutine LdSurfBands(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Character*100 :: Token, FileName, ChargeBase
  Integer       :: tlen, MaxIter, i,j,k, Kindex, NumK, NumB,Flag
  Integer :: Nnormal,kp,Psi_toUse(Mem_MapSize)
  Real  :: Tol, Kpnt(3), FracKpnt(3),ENK  ,GradEnk(3),Gperp(3)
  Real :: deltaGp(3),chk,Ctof(3,3),deltaEnk,Gmag,deltaGmag
  Type (Word_Context) :: Kpnt_WC
  Logical :: EConverged
  Type (Mem_handle), Pointer :: Psi

  bandstructure_mode=.true.
  ReCalculate_Occupancies=.false.

  Ctof=TRANSPOSE(xtal%Basis)/(2*Pi)  ! for calculating fractional k's

  Call GetNextWord(WC, Filename, tlen)
  If (W_Error /= W_OK) then
     Write(Error_Unit, *) 'LdSurfBands: Error occured while processing command!'
     Call Word_GetAndPrint(WC, Error_Unit, "LdSurfBands:")
     STOP
  End If
  Band_Name='RECIP_VECTOR'
  open(unit=Band_Out,file=Band_name,form='formatted')
  Write(Band_OUT,'("GA   ",1p3e20.12)') xtal%Recip(1:3,1)
  Write(Band_OUT,'("GB   ",1p3e20.12)') xtal%Recip(1:3,2)
  Write(Band_OUT,'("GC   ",1p3e20.12)') xtal%Recip(1:3,3)
  close(Band_OUT)

  Band_Name=Trim(FileName)//'.surfband'

  open(unit=Band_Out,file=Band_name,form='formatted',position='append')

  Call GetNumbers(WC, MaxIter)
  Call GetNumbers(WC, Tol)
  
  if (Print_Level == PRINT_VERBOSE) then
     write(LOG_Unit, *) "ldBandStruct: Relaxing Charge:  MaxIter=",MaxIter, &
          '* Tolerance=',Tol, ' * K-point File:', TRIM(FileName)
  End IF

  If (.not.HamLoaded) then
     write(Error_Unit,*) ' Error in Surface Band Calculation -- ', &
       'Must first calculate or load Hamiltonian'
     stop
  EndIf

  Call InitWord(Kpnt_WC, BAND_IN, Filename, DELIMS, COMMENT, LIT_CHAR, &
       1, INCLUDE_STR)


  !*** Create the New K-points storage space ***
  NumKpnts = 1
  BZ%TotalUniq = NumKpnts
  BZ%TotalKpnts = NumKpnts

write(log_unit,*) 'ldbandstruct -- before allocations';call flush_(Log_Unit)
  If (Associated(bz%ku)) Nullify(Bz%ku) 
  If (Associated(bz%keu)) Nullify(Bz%keu) 
  If (Associated(bz%wtuniq)) Nullify(Bz%wtuniq) 
  If (Associated(bz%weight)) Nullify(Bz%weight)
  If (Associated(xtal%wt)) Nullify(xtal%wt) 
  If (Associated(occupancy)) Nullify(occupancy) 
  If (Associated(User_Kpoints)) Nullify(User_Kpoints) 

  Allocate(BZ%Ku(3,1), xtal%Wt(1), BZ%Keu(NumBands,1), &
       BZ%WtUniq(NumBands, 1), BZ%Weight(NumBands, 1), &
       Occupancy(NumBands, NumKpnts),  User_Kpoints(3,1), stat=i)

write(log_unit,*) 'ldsurfbands -- after allocations',i;call flush_(Log_Unit)
  If (i /= 0 ) then
      write(Error_Unit,*) 'Error in LdSurfBands --' ,&    
        ' Cannot reallocate kpoint arrays ',i
      stop
  EndIf
  BZ%Kp => BZ%Ku;  
  BZ%Weight(:,1) = 1;  !***BZ%Weight(:,2) = 0;
  BZ%WtUniq = BZ%Weight

  BZ%Kp = 0
  User_Kpoints = 0

 !Define Gperp -- read in Fractional units
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) then
        Write(Error_Unit,*) 'error in ldsurfband: Need Gperp'
        stop
     Endif
 !Define Nnormal -- # number of perpendicular values of K for each K||
     Call GetNumber(Kpnt_WC, Nnormal)
     If ((W_ERROR /= W_OK) .OR. (Nnormal < 1)) then
        Write(Error_Unit,*) 'error in ldsurfband: Need Nnormal'
        stop
     Endif

     Gperp(:) = MATMUL(xtal%Recip, FracKpnt)
     DeltaGp(:) = Gperp(:)/Nnormal

  Write(Log_Unit,*) 'ldSurfBands: Nnormal = ',Nnormal
  Write(Log_Unit,*) 'ldSurfBands: Gperp   = ',Gperp  
  Write(Log_Unit,*) 'ldSurfBands: DeltaGp = ',DeltaGp
     deltaGmag=SQRT(DOT_PRODUCT(DeltaGp,DeltaGp))
  Write(Log_Unit,*) 'ldSurfBands: DeltaGmag = ',DeltaGmag
  Call Flush_(Log_Unit)

  !*** Now Get the maximum number of bands for processing
  Call GetNumbers(Kpnt_WC, NumB)      ! Maximum number of bands to calculate

  Write(Log_Unit,*) 'LdSurfBands: Processing ',NumB,' bands'
  call flush_(Log_Unit)

  Call PrepPrevious_band(NumB)

  !!!********Initialize Gradient module**************!!!
  Call InitGradEnk

  !*** Finally process the K||'s ****
  i=1
  Kpnt = 0
  Kindex=1
  Do 
     Call GetNumbers(Kpnt_WC, FracKpnt)
     If (W_ERROR /= W_OK) exit

     Write(Log_Unit,*) 'LdSurfBands: Loop=',i, ' * FracKpnt=',FracKpnt
     call flush_(Log_Unit)
     Write(Output_Unit,'("LdSurfBands: i, * FracK||=",i5,3f10.5)') i,FracKpnt
     call flush_(Output_Unit)

     Kpnt(:) = MATMUL(xtal%Recip, FracKpnt)
! Check that K|| is perpendicular to Gperp
     chk=DOT_PRODUCT(Kpnt,Gperp)
      If (ABS(chk).gt.1.e-5) then
         write(Error_Unit,*) &
            'Error in SurfBands -- K|| ,Gperp invalid',&
             Kpnt,Gperp,chk
         stop
      EndIf

      Kpnt(:) = Kpnt(:) - DeltaGp(:)/2
      Gmag = -deltaGmag/2

      Do  kp=1,Nnormal
          Kpnt(:) = Kpnt(:) + DeltaGp(:)
          Gmag = Gmag + deltaGmag
   
          User_Kpoints(:,Kindex) = MATMUL(Ctof,Kpnt)

          write(Log_Unit,*) 'SurfBand -- ', FracKpnt
          write(Log_Unit,*) 'SurfBand -- ', kp, Kpnt     
          write(Log_Unit,*) 'SurfBand -- ', kp, User_Kpoints(:,Kindex)

          Call UpdateKpoint(Kpnt)
     
          Call RelaxElectrons(MaxIter, Tol, EConverged) 

          !*** Write the info to the OUTPUT file ***
          Write(Output_Unit,'("    kp, kpnt= ", i5,3f10.5)') kp,Kpnt

          Call NewKforGradEnk(Kpnt)
          Psi_toUse=MH_Skip

          Where ((PsiInfo(:)%Kpnt > 0)) Psi_toUse=MH_toProcess
          Call Phase_Generic(Psi_toUse,Kindex)
          Call GetNextPsi(Flag,Psi)

          k = 0
          Do while (Flag > 0)
             j=Psi%Index
             ENK=PsiInfo(j)%Energy
             If (ENK <= Eigen_Max) then
                k = k + 1
                Write(Output_Unit,'("          k j ENK ERR",2i5,1p2e15.7)')&
                     k, j, ENK, PsiInfo(j)%Error
                Call CalcGradEnk(ENK, Psi%ptr, PsiInfo(j)%PDOT,GradEnk)
                deltaEnk=DOT_PRODUCT(GradEnk,DeltaGp)
                Write(Output_Unit, '(" Grad, DeltaE = ",1p4e15.7)') &
                       GradEnk, DeltaEnk
                Write(Output_Unit,'(">>>>",1p3e15.7)')Gmag,Enk,deltaEnk
                Write(Band_Out,'(5f10.5)')& 
                     Frackpnt,ENK-DeltaEnk/2,ENK+DeltaEnk/2
                Call Flush_(Band_Out)
                Call Flush_(Output_Unit)
             End If
          Call GetNextPsi(Flag,Psi)
          End Do

       EndDo    !kp
 
     i=i+1

  End Do

   Write(Log_Unit,*) 'Ending surface band calculation after processing ', &
     i-1,' k-points'
   Call Flush_(Log_Unit)
    

  Call FinishGradEnk
  Call CloseWord(Kpnt_WC)
  bandstructure_mode=.false.

  Return
End Subroutine
End Module
